# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

from cmf.monitor import schema

_BYTES_PER_MB = 1048576

# This class encapsulates the information needed to look up a metric from a
# Hadoop metrics style JSON object.
class MetricAccessor:
  def __init__(self, category, section, metric):
    self.category = category
    self.section = section
    self.metric = metric

# This class encapsulates the information needed to look up a metric from a
# JSON metrics object produced by the JMXJsonServlet.
class JmxMetricAccessor:
  def __init__(self, bean_alias, metric):
    self.bean_alias = bean_alias
    self.metric = metric

# This class encapsulates the information needed to look up a metric from a
# an Impala provided JSON metrics dictionary.
class ImpalaMetricAccessor:
  def __init__(self, metric):
    self.metric = metric

# This function builds a dictionary of metric accessors (see above) from the
# metric_source fields in the metrics schema that is structured as followed:
#
# [entity type string][version string][metric id] -> metric accessor
#
# For example, a single entry would look like the following:
#
# ['NAMENODE']['cdh3']['19178'] =
#   MetricAccessor('rpc', 'detailed-metrics' 'blockReceived_avg_time')
def build_metric_accessors(schema):
  accessors = dict()
  for metric in schema.itervalues():
    if metric.get('deprecated', 'false').lower() == 'true':
      continue
    for entity, sources in metric.get('metric_source', {}).iteritems():
      if entity not in accessors:
        accessors[entity] = dict()
      for version, source in sources.iteritems():
        if version not in accessors[entity]:
          accessors[entity][version] = dict()

        # The code below is parsing the dictionary values in the 'metric_source'
        # portion of the metric schema. Examples of those values include:
        #
        # ...
        # "NAMENODE": {
        #  "cdh3": "METRICS$$dfs.FSNamesystem.BlocksTotal",
        #  "cdh4": "JMX$$fsnamesystem::Hadoop:service=,name=FSNamesystem::BlocksTotal"
        # }
        # ...
        #
        # Here the METRICS$$ and JMX$$ indicate that the metric is either in the
        # Hadoop metrics or JMXJsonServlet format respectively. These values are
        # checked to determine what type of metric accessor to build and how the
        # remainder of the string should be parsed.
        #
        # The remainder of the METRICS$$ format looks like:
        #
        # rpc.metrics.getProtocolVersion.aboveOneSec._avg_time
        #
        # to parse this, we split on two '.'s. The first part is the category,
        # the second is the section and the remainder is the metric. Note that
        # there are metrics with '.'s in their names.
        #
        # The remainder of the JMX$$ format looks like:
        #
        # fsnamesystem::Hadoop:service=,name=FSNamesystem::BlocksTotal
        #
        # to parse this, we split on two '::'s. The first part is the bean alias,
        # the second is unused and the remainder is the metric. (The second part
        # is actually used, just not here -- it's used as a key into the alias
        # and ignored metrics maps in platform-metrics.py).
        #
        # The remainder of the IMPALA$$ format is just the key into the Impala
        # JSON metrics dictionary. An example string is:
        #
        # IMPALA$$simple-scheduler.assignments.total
        #
        if source.startswith('JMX$$'):
          # We are dealing with a JMX accessor...
          splits = source.split('JMX$$', 1)[1].split('::', 2)
          accessors[entity][version][metric['id']] = \
              JmxMetricAccessor(splits[0], splits[2])
        elif source.startswith('METRICS$$'):
          # We are dealing with a metrics page accessor...
          splits = source.split('METRICS$$', 1)[1].split('.', 2)
          accessors[entity][version][metric['id']] = \
              MetricAccessor(splits[0], splits[1], splits[2])
        elif source.startswith('IMPALA$$'):
          # We are dealing with an Impala accessor...
          accessors[entity][version][metric['id']] = \
              ImpalaMetricAccessor(source.split('$$', 1)[1])

  return accessors

# This function uses the input accessors map and adds all available metrics from
# the input metrics object to the input update. The arguments are as follows:
#
# update: a RoleUpdate
# metrics: a JsonParser object built from a Hadoop metrics sytle JSON object
# accessors: a map of metric accessors built by build_metric_accessors
# entity: an entity type string from a metric source, e.g. 'NAMENODE'
# version: a version string from a metric source, e.g. 'cdh3'
def add_metrics_metrics(update, metrics, accessors, entity, version):
  metric_accesors = accessors.get(entity, {}).get(version, {})
  for metric_id, accessor in metric_accesors.iteritems():
    default = None
    handle_duplicates = False
    # "Detailed RPC" metrics for core Hadoop (MR, HDFS) are lazily created. That
    # means that they will not be present in the metrics JSON until the first
    # time that RPC is called. We prefer inserting a 0 to this 'not present'
    # behavior.
    if accessor.category == 'rpc' and accessor.section == 'detailed-metrics':
      default = 0;
    # The task tracker does not reset its RPC metrics when it gets
    # reinitialized (see OPSAPS-3319), so we have multiple entries for
    # each metric. See the comments in get_metrics_metric for details on how
    # we deal with this.
    if entity == 'TASKTRACKER' and accessor.category == 'rpc':
      handle_duplicates = True
    update.add_metric(metric_id,
                      get_metrics_metric(metrics,
                                         accessor.category,
                                         accessor.section,
                                         accessor.metric,
                                         default,
                                         handle_duplicates))

# This function uses the input accessors map and adds all available metrics from
# the input metrics object to the input update. The arguments are as follows:
#
# update: a RoleUpdate
# metrics: a JsonParser object built from a JSON object produced by the
#          JMXJsonServlet
# accessors: a map of metric accessors built by build_metric_accessors
# entity: an entity type string from a metric source, e.g. 'NAMENODE'
# version: a version string from a metric source, e.g. 'cdh4'
def add_metrics_jmx(update, metrics, accessors, entity, version):
  metric_accesors = accessors.get(entity, {}).get(version, {})
  for metric_id, accessor in metric_accesors.iteritems():
    default = None
    # "Detailed RPC" metrics for core Hadoop (MR, HDFS) are lazily created. That
    # means that they will not be present in the metrics JSON until the first
    # time that RPC is called. We prefer inserting a 0 to this 'not present'
    # behavior.
    if accessor.bean_alias == 'rpc' or accessor.bean_alias == 'detailed_rpc':
      default = 0
    update.add_metric(metric_id,
                      get_jmx_metric(metrics,
                                     accessor.bean_alias,
                                     accessor.metric,
                                     default))

_IMPALA_ALIASES = {
   "impala-server.io-mgr.num-buffers": "impala-server.io.mgr.num.buffers",
   "impala-server.io-mgr.num-open-files": "impala-server.io.mgr.num.open-files",
   "impala-server.mem-pool.total-bytes": "impala-server.mem.pool.total-bytes",
   "impala-server.num-fragments": "impala-server.num.fragments",
   "impala-server.num-queries": "impala-server.num.queries",
   "impala-server.scan-ranges.num-missing-volume-id": "impala-server.scan_ranges.num-missing-volume-id",
   "impala-server.scan-ranges.total": "impala-server.scan.ranges.total",
   "simple-scheduler.local-assignments.total": "simple-scheduler.local.assignments.total"
}

# This function uses the input accessors map and adds all available metrics from
# the input metrics object to the input update. The arguments are as follows:
#
# update: a RoleUpdate
# metrics: a JsonParser object built from a JSON object produced by Impala
# accessors: a map of metric accessors built by build_metric_accessors
# entity: an entity type string from a metric source, e.g. 'IMPALA'
# version: a version string from a metric source, e.g. 'cdh4'
def add_metrics_impala(update, metrics, accessors, entity, version):
  if metrics is None or metrics.json is None:
    return
  metric_accesors = accessors.get(entity, {}).get(version, {})
  for metric_id, accessor in metric_accesors.iteritems():
    value = None
    if accessor.metric in metrics.json:
      value = metrics.json.get(accessor.metric)
    # Impala adjusted the names of some of their metrics just before CM 4.5.
    # To be nice, we'll try looking up metrcs under their old names if they
    # are not present under their current names.
    if value == None and accessor.metric in _IMPALA_ALIASES:
      value = metrics.json.get(_IMPALA_ALIASES[accessor.metric])
    if value is not None:
      # The Impala JSON exposes its numbers as strings, e.g. "1" instead of 1.
      # This means we have to coerce its metric values into number formats. We
      # use float below because all our metrics are ultimately just stored as
      # doubles anyway. We will undo this change when Impala adjusts their JSON
      # printing code, but we are waiting for an opportunity for a compatability
      # break to do so. See OPSAPS-11201 for more detail.
      update.add_metric(metric_id, float(value))

# Utility function to get a metric value from a JsonParser built from a Hadoop
# metrics style JSON object. The JSON structure looks something like:
#
# {
# ...
#   "jvm":{
#      "metrics":[
#         [
#            {
#               "hostName":"centos62-25.ent.cloudera.com",
#               ...
#            },
#            {
#               "gcCount":41066,
#               ...
#            }
#         ]
#      ]
#   },
#   ...
# }
# 
# We refer to the 'jvm' key as the category, the 'metrics' key as section, and
# 'gcCount' as the metric.
#
# Note that this format can be a little tricky. For example, in OPSAPS-3319 we
# have entries for the same metrc in different sections because the tasktracker
# doesn't clear out the old metrics when it gets re-inited. Callers can indicate
# to handle this with the 'handle_duplicates' parameter. This flag works as
# follows:
# 1. For most metrics sum together all the values. Since most metrics are
# counters this just works
# 2. For metrics ending in _avg_time get the first non-zero value. This works
# because those are time varying rate metrics which get reset every minute so
# they will only be non-zero for active metrics (if two happen to be active
# in the same minute then we'll combine them, but something weird was happening
# then anyway, so let's no worry about it)
def get_metrics_metric(metrics, category, section, metric, default,
    handle_duplicates=False):
  if metrics is None or metrics.json is None:
    return default
  category_dict = metrics.json.get(category, None)
  if category_dict is None:
    return default
  section_list = category_dict.get(section, None)
  if section_list is None:
    return default
  total = 0
  for s in section_list:
    for metric_dictionary in s:
      if metric in metric_dictionary:
        metric_value = metric_dictionary.get(metric)
        if handle_duplicates:
          total = total + metric_value
        else: 
          return metric_value
  if handle_duplicates:
    return total
  return default

# Utility function to get a metric value from a JsonParserJmx built from a JSON
# object produed by the JMXJsonServlet.  The JSON structure looks something
# like:
#
# {
# ...
#  }, {
#    "name" : "Hadoop:service=DataNode,name=JvmMetrics",
#    ...
#    "GcCount" : 82938,
#    ...
#  }, {
# ...
# }
# 
# See JsonParserJmx for how the above is parsed to understand the access code
# in this function.
def get_jmx_metric(metrics, bean, metric, default=None):
  if metrics is None:
    return default
  if not hasattr(metrics, bean):
    return default
  return getattr(metrics, bean).get(metric, default)

# This is a utility method to add some metrics derived from raw JVM metrics that
# are in the Hadoop metrics style.
def add_derived_jvm_metrics_metrics(update, metrics):
  update.add_metric(schema.JVM_TOTAL_THREADS,
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsBlocked',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsNew',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsRunnable',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsTerminated',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsTimedWaiting',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsWaiting',
                                       0))

# This is a utility method to add some metrics derived from raw JVM metrics that
# are in the JMX metrics style.
def add_derived_jvm_metrics_jmx(update, metrics):
  # The JMX version of the JvmMetrics does not cotain the max memory of the JVM,
  # so we calculate it from a different bean.
  if metrics.java_memory is not None:
    update.add_metric(schema.JVM_MAX_MEMORY_MB,
                      metrics.java_memory.HeapMemoryUsage.max / float(_BYTES_PER_MB))

  if metrics.jvm is not None:
    update.add_metric(schema.JVM_TOTAL_THREADS,
                      metrics.jvm.get('ThreadsBlocked', 0) +
                      metrics.jvm.get('ThreadsNew', 0) +
                      metrics.jvm.get('ThreadsRunnable', 0) +
                      metrics.jvm.get('ThreadsTerminated', 0) +
                      metrics.jvm.get('ThreadsTimedWaiting', 0) +
                      metrics.jvm.get('ThreadsWaiting', 0))

class ParseError(Exception):
  pass

def get_beans(dic, names):
  """
  Given a Hadoop or Cloudera Manager JMX json dict, returns a map of beans found
  matching the input set of name patterns as a dictionary from name pattern -> bean
  as a dictionary. May raise ParseError.

  Returns a DictWithAttr.
  """
  beans = dic.get('beans')
  if beans is None:
    err = "Cannot find beans list in input json"
    raise ParseError(err)

  if not isinstance(beans, list):
    err = "Expecting a beans list. Got %r" % (beans, )
    raise ParseError(err)
  
  ret = {}
  for bean in beans:
    d = DictWithAttr(bean)
    name = d.get('name')
    for pattern in names:
      if pattern.match(name):
        ret[pattern] = d
        break
    if len(ret.keys()) == len(names):
      break

  return ret

class DictWithAttr(dict):
  """
  A wrapper class to allow attribute access for dictionary-like
  objects (e.g. json objects).
  """
  def __init__(self, dictionary, raise_when_not_found=False):
    try:
      dict.__init__(self, dictionary)
    except:
      raise ParseError("Cannot initialize dictionary with %r" % (dictionary,))
    self._raise = raise_when_not_found

  def __getattr__(self, attr):
    try:
      res = self[attr]
      if isinstance(res, dict):
        # Recursively wrap the result if it's also a dictionary
        return DictWithAttr(res)
      return res
    except KeyError:
      if self._raise:
        raise AttributeError(attr)
      else:
        return None
